#!/usr/bin/env ruby

# Do not buffer output
STDOUT.sync = true

##############################################################################
# Environment Configuration
##############################################################################
ONE_LOCATION=ENV["ONE_LOCATION"]
USER=ENV["user"]
TIMEOUT=20
MAXWAIT=60
INTERVAL=1
ONEROOT="/var/lib/one"

# oneadmin user flag value
USERFLAG=-2

if !ONE_LOCATION
    RUBY_LIB_LOCATION="/usr/lib/one/ruby"
else
    RUBY_LIB_LOCATION=ONE_LOCATION+"/lib/ruby"
end

$: << RUBY_LIB_LOCATION

##############################################################################
# Required libraries
##############################################################################
require 'opennebula'
require 'optparse'
require 'find'
require 'fileutils'
require 'syslog/logger'
require 'time'
require 'pathname'
require 'yaml'

include OpenNebula

LOGGER   = Syslog::Logger.new("HapyBackup")

class HapyImage
  attr_reader :id, :source, :vms, :running_vms
  def initialize(imgObject)
    @id = imgObject.to_hash['IMAGE']['ID']
    @image = imgObject
    @info  = imgObject.to_hash['IMAGE']
    @owner = @info['UNAME']
    @source = @info['SOURCE']
    @vms = @info['VMS']
    @running_vms = @info['RUNNING_VMS']
  end

  def name
    return @image.name
  end

  def persistent
    if @info['PERSISTENT'] == "1"
      return true
    else
      return false
    end
  end

  def remove
    begin
      @image.delete
    rescue
      @image.info
      begin
        @image.delete
      rescue
        return false
      end
      return true
    end
  end

  def status
    @image.info
    @info = @image.to_hash['IMAGE']
    return @image.short_state_str
  end

  def show
    print("#{self.name}\n")
    print("\tOwner: #{self.owner}\n")
    print("\tPersistent: #{self.persistant}\n")
    print("\tSource: #{self.source}\n")
    print("\tvms: #{self.vms}\n")
    print("\trunning_vms: #{self.running_vms}\n")
  end
end

class HapyDisk
    attr_reader :id,:source,:path
    def initialize(dsk)
        @id = dsk["DISK_ID"].to_i
        @imageID = dsk["IMAGE_ID"]
        @imageName = dsk["IMAGE"]
        @source = dsk["SOURCE"]
        @datastore = dsk["DATASTORE"]
        @persistent = dsk["PERSISTENT"]
    end

    def is_persistent?()
        if @persistent == "YES"
            return true
        end
        return false
    end

    def show()
        puts("   Disk ID : #{@id}")
        puts("  Image ID : #{@imageID}")
        puts("Image Name : #{@imageName}")
        puts(" Datastore : #{@datastore}")
        puts("    Source : #{@source}")
    end
end

class HapyVM
    attr_reader :id, :info, :name, :disks, :images, :labels
    def initialize(vm, images)
        @images = self._update_image_pool(images)
        @info = vm.to_hash
        @id = @info["VM"]["ID"]
        @name = @info["VM"]["NAME"]
        @labels = self._get_labels(@info["VM"]["USER_TEMPLATE"]["LABELS"])
        @disks = self._get_disks(@info["VM"]["TEMPLATE"]["DISK"])
        @deployment_files = self._locate_deployement()
        @vm_path = self._get_vm_files_path()
        @context_disk_id = @info["VM"]["TEMPLATE"]["CONTEXT"]["DISK_ID"]
        @vm = vm
    end

    def _get_labels(lbs)
        if lbs
            return lbs.split(",")
        end
        return []
    end

    def _get_vm_files_path()
        ds_id = 0
        history = @info["VM"]["HISTORY_RECORDS"]["HISTORY"]
        if history.is_a?(Hash)
            ds_id = @info["VM"]["HISTORY_RECORDS"]["HISTORY"]["DS_ID"]
        elsif history.is_a?(Array)
            ds_id = @info["VM"]["HISTORY_RECORDS"]["HISTORY"][-1]["DS_ID"]
        else
            return nil
        end
        return "#{ONEROOT}/datastores/#{ds_id}/#{@id}"
    end

    def _update_image_pool(pool)
        newpool = {}
        newpool.store("pool",pool["pool"])
        pool["pool"].info
        pool["pool"].each do |img|
            infoid = img.to_hash['IMAGE']['ID']
            newpool.store(infoid, HapyImage.new(img))
        end
        return newpool
    end

    # Aim: Return a list of disks (HapyDisk)
    # Param: Virtual machine template disk information from VM/TEMPLATE/DISK
    # Note: This param ca be a hash or a list of hash so we need check the information type
    #       in order to provide the correct disk information.
    # Return : Allways a list of HapyDisk objects (even for on disk)
    def _get_disks(dsinfo)
        disks = []
        if dsinfo.is_a?(Array)
            dsinfo.each do |ds|
                disks.append(HapyDisk.new(ds))
            end
        else
            disks.append(HapyDisk.new(dsinfo))
        end
        
        return disks
    end

    def _locate_deployement()
        depl = []
        patt = ".*\/datastores\/.*\/#{@id}\/deployment.[0-9]"
        Find.find(ONEROOT) do |path|
            if path =~ /#{patt}/
                depl.append(path)
            end
        end
        return depl
    end

    def state
        @vm.info
        return @info["VM"]["STATE"]
    end

    def _wait_for_it(st)
        cnt = 0
        sleep(1)
        while @vm.state != st.to_i
            cnt += 1
            sleep(1)
            if cnt == 120
                break
            end
        end
    end

    # Aim: Backup a virtual machine 
    # Param: none
    # Return: backup status as an int (0 success 1 or more failure)
    def backup()
        bck_root = "#{BACKUP_ROOT}/one" 
        dest_dir = ""
        vm_path = ""

        # Backup deployement files
        # The files are saved to ONEROOT with the datastore directory structure
        @deployment_files.each do |dpl|
            # get the full original file path
            pth=Pathname.new(File.expand_path(dpl))
            # remove the oneroot from it to keep the datastore directory structure 
            rlv=pth.relative_path_from(ONEROOT)
            # set the file destination directory with the datastore directory structure
            dest_dir="#{bck_root}/#{File.dirname(rlv)}"
            if ! File.exists?(dest_dir)
                FileUtils.mkdir_p(dest_dir)
            end
            # Backup the deployement file
            FileUtils.copy_file(dpl, "#{dest_dir}/#{File.basename(dpl)}", preserve='true')
        end


        # Backup symlinks
        Find.find(@vm_path) do |path|
            if File.symlink?(path)
                lk="#{dest_dir}/#{File.basename(path)}"
                if ! File.exists?(lk)
                    FileUtils.copy_entry(path,lk)
                end
            end
        end


        # Backup OS disks
        # Create backup image 
        @disks.each do |vmds|
            initial_state = self.state
            bck_img_id = @vm.disk_saveas(vmds.id,"BCK-#{@vm.id}-disk-#{vmds.id}-#{Time.now.to_i}")
            if OpenNebula.is_error?(bck_img_id)
                LOGGER.error(bck_img_id.message)
                next
            end

            self._wait_for_it(initial_state)

            #self._wait_for_it(img_pool, bck_img_id)
            to_clean = []
            @images = self._update_image_pool(@images)
            @images.each do |key, img|
                if key == "pool"
                    img.info
                    next
                end

                if img.id.to_i == bck_img_id
                    wait = 0
                    while img.status != 'rdy'
                        images["pool"].info
                        wait += INTERVAL
                        sleep(INTERVAL)
                        if wait >= MAXWAIT
                            break
                        end
                    end

                    imFile = img.source

                    dstFile = ""
                    oriFile = ""
                    if ! vmds.path
                        oriFile = vmds.source
                    else
                        oriFile = vmds.path
                    end
                    pth = Pathname.new(File.expand_path(oriFile))
                    rlv = pth.relative_path_from(ONEROOT)

                    if vmds.is_persistent?()
                        dstFile="#{bck_root}/#{File.dirname(rlv)}/#{File.basename(oriFile)}"
                    else
                        pth = Pathname.new(File.expand_path(@vm_path))
                        rlv = pth.relative_path_from("#{ONEROOT}/datastores")
                        dstFile="#{bck_root}/datastores/#{rlv}/disk.#{vmds.id}"
                    end

                    sum = Digest::SHA1.file(imFile)
                    if not copyImage(imFile, dstFile, sum, compress=false)
                        LOGGER.error("Backup image of vm #{@vm.name} failed on disk copy #{imFile} -> #{dstFile}")
                        next 
                    else
                        to_clean.append(img)
                    end
                end
            end
            to_clean.each do |img|
                oldID=img.id
                img.remove
                @images = self._update_image_pool(@images)
                if @images["#{oldID}"]
                    img.remove
                end
                sleep(1)
            end

        end
        # Backup context disk
        if @context_disk_id
            pth = Pathname.new(File.expand_path(@vm_path))
            rlv = pth.relative_path_from("#{ONEROOT}")
            imFile = "#{@vm_path}/disk.#{@context_disk_id}"
            dstFile = "#{bck_root}/#{rlv}/disk.#{@context_disk_id}"
            sum = Digest::SHA1.file(imFile)
            if not copyImage(imFile, dstFile, sum, compress=false)
                return 1
            end
        end

        return 0
    end

    def show()
        puts "Virtual machine: "
        puts "    ID: #{@id}"
        puts "  NAME: #{@name}"
        puts " STATE: #{@state}"
        puts "LABELS:"
        @labels.each do |lb|
            puts("  - #{lb}")
        end
        puts "  DISK: "
        @disks.each do |ds|
            ds.show()
        end
    end
end


def copyImage(source, dest, orisum, compress=false)
    #if File.symlink?(source)
    #    source=File.readlink(source)
    #end
    if ! File.exists?(File.dirname(dest))
        FileUtils.mkdir_p(File.dirname(dest))
    end
    if compress
       res = compressImage(source, dest)
       if res == 0
          LOGGER.info("Image #{source} saved to #{dest}")
          return true
       else
          LOGGER.error("Saving #{source} to #{dest} failed")
          return false
       end
    else
       dest.delete("\s")
       LOGGER.info("cp #{source} #{dest}")
       FileUtils.cp(source, dest)
       newsum = Digest::SHA1.file(dest)
       if orisum == newsum
          LOGGER.info("Image #{source} saved to #{dest}")
          return true
       else
          LOGGER.error("Saving #{source} to #{dest} failed")
          return false
       end
    end
end

#
# NAME: compressImage
# PARAM:
#    - source: file to compress
# AIM: Copy Disk file into the backup directory
#
def compressImage(source, dest=nil)
  if not dest
     dest = "#{source}.z"
  end
  LOGGER.info("Compression of #{source} into #{dest}")
  system("qemu-img convert -O qcow2 -c '#{source}' '#{dest}'")
  return $?.exitstatus
end

options = {:creds => nil, :action => nil, :endpoint => nil,
           :timeout => nil}

parser = OptionParser.new do|opts|
    opts.banner = "Usage: #{File.basename(__FILE__)} [options]"

    opts.on('-C', '--config-file file', 'Configuration file') do |value|
        options[:config_file] = value;
    end

    opts.on('-c', '--creds file', 'Crediential file') do |value|
        options[:creds] = value;
    end

    opts.on('-e', '--end-point url', 'End point URL') do |value|
        options[:endpoint] = value;
    end

    opts.on('-d', '--backup-directory dir', 'Backup directory') do |value|
        options[:backup_dir] = value;
    end

    opts.on('-T', '--backup-tag tag', 'Tag who marks vm for backup (ex: Backup)') do |value|
        options[:backup_only_tagged] = "oui"
        options[:backup_tag] = value;
    end

    opts.on('-t', '--timeout timeout', 'Timeout for opennebula connection') do |value|
        options[:timeout] = value.to_i;
    end

    opts.on('-u', '--backup-unused', 'Backup images not used by any VM') do |value|
        options[:backup_unused] = "oui";
    end

    opts.on('-w', '--wait timeout', 'Wait for action ends') do |value|
        options[:wait] = value.to_i
    end

    opts.on('-h', '--help', 'Displays Help') do
        puts opts
        exit
    end

end

parser.parse!

if not options[:config_file]
    options[:config_file] = "/etc/one/hapybck.conf"
end

CONFIG = YAML::load_file(options[:config_file])

# OpenNebula credentials
if not options[:creds]
    options[:creds] = "/var/lib/one/.one/one_auth"
end

if not options[:endpoint]
    ip = CONFIG[:endpoint] 
    options[:endpoint] = "http://#{ip}/RPC2"
end

if not options[:backup_dir]
    options[:backup_dir] = CONFIG[:backup_dir]
end

if not options[:backup_tag]
    options[:backup_tagged] = CONFIG[:backup_only_tagged]
    options[:backup_tag] = CONFIG[:backup_tag]
end

if not options[:backup_unused]
    options[:backup_unused] = CONFIG[:backup_unused]
end

begin
    File.readlines(options[:creds]).each do |line|
        CREDENTIALS = line
    end
rescue
    puts("#{options[:creds]}: Problem loading credentials, check if file exists.")
    exit(-1)
end

exit_code = 0
begin
    BACKUP_ROOT = options[:backup_dir]
    puts(BACKUP_ROOT)
    client = Client.new(CREDENTIALS, options[:endpoint])

    vm_pool = VirtualMachinePool.new(client, USERFLAG)
    # Try to load vm pool infos from OpenNebula until timeout expires
    rc = vm_pool.info
    cnt = 0
    while OpenNebula.is_error?(rc)
        if cnt == options[:timeout]
            puts rc.message
            exit(-1)
        end
        rc = vm_pool.info
        sleep(1)
        cnt += 1
    end

    backup_tag = options[:backup_tag]
    backup_tagged = options[:backup_only_tagged]
    backup_unused = options[:backup_unused]

    # Backup unsed images (if needed) and used non persistant images
    images = {}
    img_pool = ImagePool.new(client, USERFLAG)
    rc = img_pool.info
    cnt = 0
    while OpenNebula.is_error?(rc)
        if cnt == 10
            puts rc.message
            exit 34
        end
        rc = img_pool.info
        sleep(1)
        cnt += 1
    end

    # Store images in a clean hash of HapyImage objects
    images.store("pool", img_pool)
    img_pool.each do |img|
      infoid = img.to_hash['IMAGE']['ID']
      images.store(infoid, HapyImage.new(img))
    end

    puts("Saving unused (if needed) and non persistent images")
    images.each do |key, img|
        if key == "pool"
            img.info
            next
        end
        imgFile = img.source
        if img.status == 'err' or img.status ==  'dele'
            puts("Image #{img.id} #{img.name} state is [#{img.status}]... skipping backup")
            next
        end
        if img.vms.empty? and backup_unused == "oui"
            pth = Pathname.new(File.expand_path(imgFile))
            rlv = pth.relative_path_from(ONEROOT)
            destFile = "#{BACKUP_ROOT}/one/#{rlv}"
            sum = Digest::SHA1.file(imgFile)
            if not copyImage(imgFile, destFile, sum, compress=false)
                return 1
            end
        elsif ! img.persistent
            pth = Pathname.new(File.expand_path(imgFile))
            rlv = pth.relative_path_from(ONEROOT)
            destFile = "#{BACKUP_ROOT}/one/#{rlv}"
            sum = Digest::SHA1.file(imgFile)
            if not copyImage(imgFile, destFile, sum, compress=false)
                return 1
            end
        end
    end


    hapyvms = []

    vm_pool.each do |vm|
        vm.info
        hapyvms.append(HapyVM.new(vm,images))
    end

    hapyvms.each do |vm|
        LOGGER.info("Working on VM #{vm.id} - #{vm.name}")
        if backup_tagged == "oui"
            if vm.labels.include?(backup_tag)
                vm.backup()
            else
                LOGGER.info("Ignoring untagged VM #{vm.id} - #{vm.name}")
            end
        else
            vm.backup()
        end
    end


rescue Exception => e
    puts e.message
    puts e.backtrace
    exit(-1)
end
exit(exit_code)