#!/usr/bin/env ruby

# Do not buffer output
STDOUT.sync = true

##############################################################################
# Environment Configuration
##############################################################################
ONE_LOCATION=ENV["ONE_LOCATION"]
USER=ENV["user"]
TIMEOUT=20
MAXWAIT=60
INTERVAL=1
ONEROOT="/var/lib/one"

# oneadmin user flag value
USERFLAG=-2

if !ONE_LOCATION
    RUBY_LIB_LOCATION="/usr/lib/one/ruby"
else
    RUBY_LIB_LOCATION=ONE_LOCATION+"/lib/ruby"
end

$: << RUBY_LIB_LOCATION

##############################################################################
# Required libraries
##############################################################################
require 'opennebula'
require 'optparse'
require 'find'
require 'fileutils'
require 'syslog/logger'
require 'time'
require 'pathname'
require 'yaml'

include OpenNebula

LOGGER   = Syslog::Logger.new("HapyRestore")

class HapyVM
    attr_reader :id, :info, :name, :disks, :images, :labels
    def initialize(vm, images)
        if images
            @images = self._update_image_pool(images)
        else
            @images = images
        end
        @info = vm.to_hash
        @id = @info["VM"]["ID"]
        @name = @info["VM"]["NAME"]
        @vm = vm
    end

    def _wait_for_it(st)
        cnt = 0
        sleep(1)
        while self.status != st
            cnt += 1
            sleep(1)
            if cnt == 120
                break
            end
        end
    end

    def state
        @vm.info
        return @info["VM"]["STATE"]
    end

    def status
        @vm.info
        return @vm.status
    end

    def poweroff(hard=false)
        @vm.info
        @vm.poweroff(hard=true)
        self._wait_for_it("poff")
    end

    def resume
        @vm.info
        @vm.resume
        self._wait_for_it("runn")
    end
end

def copyImage(source, dest, orisum, compress=false)
    #if File.symlink?(source)
    #    source=File.readlink(source)
    #end
    if ! File.exists?(File.dirname(dest))
        FileUtils.mkdir_p(File.dirname(dest))
    end
    if compress
       res = compressImage(source, dest)
       if res == 0
          LOGGER.info("Image #{source} saved to #{dest}")
          return true
       else
          LOGGER.error("Saving #{source} to #{dest} failed")
          return false
       end
    else
       dest.delete("\s")
       LOGGER.info("cp #{source} #{dest}")
       FileUtils.cp(source, dest)
       newsum = Digest::SHA1.file(dest)
       if orisum == newsum
          LOGGER.info("Image #{source} saved to #{dest}")
          return true
       else
          LOGGER.error("Saving #{source} to #{dest} failed")
          return false
       end
    end
end

def getDest(filename)
    pth=Pathname.new(File.expand_path(filename))
    rlv=pth.relative_path_from(BACKUP_ROOT)
    return "#{ONEROOT}/#{rlv}"
end

def restore_vm(vm, id,path)
    bck_dir = File.dirname(path)
    puts("Restoring vm #{id} from #{bck_dir}")
    vm.info
    vm.poweroff
    files = Find.find(bck_dir)
    files.each do |file|
        if File.directory?(file)
            next
        end
        if File.symlink?(file)
            # Follow the link
            lkDest = File.readlink(file)
            # Get the file shortname
            lkDestFile = File.basename(lkDest)

            # Find the backuped file
            bckFile = ""
            patt = ".*#{lkDestFile}"
            Find.find(BACKUP_ROOT) do |path|
                if path =~ /#{patt}/
                    bckFile = path
                end
            end

            # Set the destination file for recover
            dest=getDest(bckFile)
            #Recover the file
            sum = Digest::SHA1.file(bckFile)
            if not copyImage(bckFile, dest, sum, compress=false)
                LOGGER.error("VM #{id} #{vm.name} : Recover of #{bckFile} to #{dest} failed !")
                puts("ERROR on #{bckFile} recovery from VM #{id}")
            end
            
            begin
                ldest=getDest(file)
                if File.symlink?(ldest)
                    if File.readlink(file) == File.readlink(ldest)
                        next
                    else
                        FileUtils.remove_entry(ldest)
                        FileUtils.copy_entry(file,ldest)
                    end
                else
                    LOGGER.error("Source file [#{file}] is a symlink but destination [#{ldest}] is not")
                    puts("ERROR - Source file [#{file}] is a symlink but destination [#{ldest}] is not")
                end
            rescue Exception => e
                LOGGER.error(e.message)
                puts(e.message)
            end
        else
            # Set the destination file for recover
            dest=getDest(file)
            #Recover the file
            sum = Digest::SHA1.file(file)
            if not copyImage(file, dest, sum, compress=false)
                LOGGER.error("VM #{id} #{vm.name} : Recover of #{file} to #{dest} failed !")
                puts("ERROR on #{file} recovery from VM #{id}")
            end
        end
    end
    vm.resume
end

options = {:creds => nil, :action => nil, :endpoint => nil,
           :timeout => nil}

parser = OptionParser.new do|opts|
    opts.banner = "Usage: #{File.basename(__FILE__)} [options]"

    opts.on('-C', '--config-file file', 'Configuration file') do |value|
        options[:config_file] = value;
    end

    opts.on('-c', '--creds file', 'Crediential file') do |value|
        options[:creds] = value;
    end

    opts.on('-e', '--end-point url', 'End point URL') do |value|
        options[:endpoint] = value;
    end

    opts.on('-d', '--backup-directory dir', 'Backup directory') do |value|
        options[:backup_dir] = value;
    end

    opts.on('-T', '--backup-tag tag', 'Tag who marks vm for backup (ex: Backup)') do |value|
        options[:backup_only_tagged] = "oui"
        options[:backup_tag] = value;
    end

    opts.on('-t', '--timeout timeout', 'Timeout for opennebula connection') do |value|
        options[:timeout] = value.to_i;
    end

    opts.on('-u', '--backup-unused', 'Backup images not used by any VM') do |value|
        options[:backup_unused] = "oui";
    end

    opts.on('-m', '--machines vmid', "VM ID list (one or more ',' separated)") do |value|
        options[:machines] = value.split(',')
    end

    opts.on('-w', '--wait timeout', 'Wait for action ends') do |value|
        options[:wait] = value.to_i
    end

    opts.on('-h', '--help', 'Displays Help') do
        puts opts
        exit
    end

end

parser.parse!

if not options[:config_file]
    options[:config_file] = "/etc/one/hapybck.conf"
end

CONFIG = YAML::load_file(options[:config_file])

# OpenNebula credentials
if not options[:creds]
    options[:creds] = "/var/lib/one/.one/one_auth"
end

if not options[:endpoint]
    ip = CONFIG[:endpoint] 
    options[:endpoint] = "http://#{ip}/RPC2"
end

if not options[:backup_dir]
    options[:backup_dir] = CONFIG[:backup_dir]
end

if not options[:backup_tag]
    options[:backup_tagged] = CONFIG[:backup_only_tagged]
    options[:backup_tag] = CONFIG[:backup_tag]
end

if not options[:backup_unused]
    options[:backup_unused] = CONFIG[:backup_unused]
end

if not options[:machines]
    options[:machines] = ["all"]
end

begin
    File.readlines(options[:creds]).each do |line|
        CREDENTIALS = line
    end
rescue
    puts("#{options[:creds]}: Problem loading credentials, check if file exists.")
    exit(-1)
end

exit_code = 0
begin
    BACKUP_ROOT = File.join(options[:backup_dir],"one")
    client = Client.new(CREDENTIALS, options[:endpoint])

    hapy_vms = {}
    vm_pool = VirtualMachinePool.new(client, USERFLAG)
    # Try to load vm pool infos from OpenNebula until timeout expires
    rc = vm_pool.info
    cnt = 0
    while OpenNebula.is_error?(rc)
        if cnt == options[:timeout]
            puts rc.message
            exit(-1)
        end
        rc = vm_pool.info
        sleep(1)
        cnt += 1
    end

    vm_pool.each do |vm|
        id = vm.to_hash["VM"]["ID"]
        hapy_vms.store(id, HapyVM.new(vm,nil))
    end

    if options[:machines][0] ==  "all"
        patt = ".*\/datastores\/.*\/.*\/deployment.[0-9]"
        vmid = {}
        Find.find(File.join(BACKUP_ROOT)) do |path|
            if path =~ /#{patt}/
                id = path.split("/")
                vmid.store(id[-2],path)
            end
        end
        idx = 0
        vmid.each do |id,path|
            restore_vm(hapy_vms["#{id}"],id,path)
        end
    else
        options[:machines].each do |machine|
            patt = ".*\/datastores\/.*\/#{machine}/deployment.[0-9]"
            bck_path = ""
            Find.find(File.join(BACKUP_ROOT)) do |path|
                if path =~ /#{patt}/
                id = path.split("/")
                    bck_path = path
                end
            end
            restore_vm(hapy_vms["#{machine}"],machine, bck_path)
        end
    end
rescue Exception => e
    puts e.message
    puts e.backtrace
    exit(-1)
end
exit(exit_code)